// objlist.c
#include <malloc.h>
#include <stdlib.h> // exit()

#include "runc.h"

typedef struct
{
  const char* id; // must be first field in struct for address calc
  long n_items;
}OBJLIST_HDR;

OBJLIST_HDR* base;
OBJLIST* current;
int current_ctr;

typedef struct
{
  const char* id;
  long n_items;
  char* s1;
  char* s2;
  char* s3;
  char* s4;
}OBJLIST_VIEW;

OBJLIST_VIEW** view = (OBJLIST_VIEW**)&base;

// returns index of first item (always 0 if any items in the array, else -1.
int objlist_first(OBJLIST* plist)
{
  if(objlist_count(plist))
    return 0;
  else 
    return -1;
}

// returns index of last item (count -1), returns -1 if none.
int objlist_last(OBJLIST* plist)
{
  return objlist_count(plist) - 1;
}

// for detecting valid objlist derivatives
static const char* objlist_ID = "objlist";

// for displaying errors and warnings, 1 = warn, 2 = fatal.
static void objlist_perror(int error_level)
{
  const char* msg = "%s [%s]: invalid object pointer\n";
  if(error_level == 2)
  {
    fprintf(stderr, msg, "Fatal", objlist_ID);
    exit(1);
  }
  else if(error_level == 1)
    fprintf(stderr, msg, "Warning", objlist_ID);
}

// for detecting imbalances in enter/leave pairs
static void objlist_nestError()
{
  fprintf(stderr, "Error: enter/leave in objlist imbalanced\n");
  exit(1);
}

// quick checks if new object pointer and is valid, returns 0 if ok
static int objlist_enter(OBJLIST* any, int error_level)
{
  if(current_ctr == 0)
    current = any;
  else if(current != any)
    objlist_nestError();
  current_ctr++;
  do
  {
    if(!any) break;
    if(!*any) break;
    base = ((OBJLIST_HDR*)(*any))-1;
    if(base->id != objlist_ID) break;
    return /* valid = */ 1;
  }while(0);
  // what to do if not valid
  if(error_level) objlist_perror(error_level);
  // else
  return 0; // false;
}

// unnests from objlist_enter() call
static void objlist_leave(OBJLIST* any)
{
  if(current != any)
    objlist_nestError();
  
  current_ctr--;
  if(current_ctr == 0) 
    current = 0;
}

// returns a pointer to the first (null) object in a new list.
OBJLIST objlist_new()
{
  base = (OBJLIST_HDR*)malloc(sizeof(OBJLIST_ITEM*) + sizeof(OBJLIST_HDR));
  base->id = objlist_ID;
  base->n_items = 0;
  base[1].id = 0;
  return (OBJLIST_ITEM**)&base[1].id;
}


// calls destructor on items and deletes the base
void objlist_delete(OBJLIST* plist)
{
  objlist_enter(plist, 2);
  OBJLIST_ITEM** list = *plist;
  base = ((OBJLIST_HDR*)list) - 1;
  free(base);
  *plist = 0;
  objlist_leave(plist);
}

// returns item at index
OBJLIST_ITEM* objlist_get(OBJLIST* plist, int index)
{
  if(index >= objlist_count(plist))
    return 0;
  if(index < 0)
    return 0;
  return (*plist)[index];
}

int objlist_count(OBJLIST* plist)
{
  return ((OBJLIST_HDR*)*plist)[-1].n_items;
}


static void objlist_setCount(OBJLIST* plist, int n)
{
  ((OBJLIST_HDR*)*plist)[-1].n_items = n;
}


// removes items, does not resize memory nor delete the base
void objlist_clear(OBJLIST* plist)
{
  int i; // loop index
  objlist_enter(plist, 2);
  OBJLIST_ITEM** list = *plist;
  for(i = 0; i < base->n_items; i++)
    list[i] = 0;
  base = ((OBJLIST_HDR*)list) -1;
  base->n_items = 0;
  *plist = list;
  objlist_leave(plist);
}


void objlist_append(OBJLIST* plist, OBJLIST_ITEM* o)
{
  objlist_enter(plist, 2); // 2 = fatal
  OBJLIST_ITEM** list = *plist;
  base = ((OBJLIST_HDR*)list) - 1;
  // add two below for new item plus terminator below
  int next_size = sizeof(OBJLIST_HDR) + (sizeof(OBJLIST_ITEM*) * (base->n_items + 2) + sizeof(sizeof(OBJLIST_HDR)));
  base = (OBJLIST_HDR*)realloc(base, next_size);
  list = (OBJLIST_ITEM**)&base[1].id;
  list[base->n_items] = o;
  base->n_items++;
  list[base->n_items] = 0; // null last object, always
  *plist = list;
  objlist_leave(plist);
}

// removes (and frees) item at idx if any. Does not resize the list
void objlist_remove(OBJLIST* plist, int idx)
{
  int i; // loop index
  objlist_enter(plist, 2);
  OBJLIST_ITEM** list = *plist;
  int n_items = objlist_count(plist);
  if((idx < 0) || (idx > n_items))
  {
    objlist_leave(plist);
    return;
  }
  
  for(i = idx; i < n_items; i++)
    list[i] = list[i+1];
  
  objlist_setCount(plist, n_items - 1);
  objlist_leave(plist);
}


// inserts new item at idx, or end or at beginning if out of range
void objlist_insert(OBJLIST* plist,  int idx, OBJLIST_ITEM* o)
{ 
  int i; // loop index 
  OBJLIST list = *plist;
  // easiest to append and then to move
  objlist_append(plist, (void*)o);
  int n_items = objlist_count(plist);
  
  // check range 
  if(idx < 0)
    idx = 0;
  else if(idx >= n_items)
    idx = n_items - 1;
  
  if(idx == n_items - 1) 
    return;
  
  // grab the appended object, then shift the list up
  // and put the new object in at idx.
  void* tmp = (void*)list[n_items - 1];
  
  // point to last and copy prev to current and decrement
  for(i = n_items - 1; i > idx; i--)
    list[i] = list[i-1];
  list[idx] = tmp;
}

// swap two items only if both indexes are valid
void objlist_swap(OBJLIST* plist, int idx1, int idx2)
{
  OBJLIST list = *plist;
  int n = objlist_count(plist);
  if(idx1 < 0 || idx1 >= n)
    return;
  if(idx2 < 0 || idx2 >= n)
    return;
  OBJLIST_ITEM* tmp = list[idx1];
  list[idx1] = list[idx2];
  list[idx2] = tmp;
}

